"""Wrapper for fastKDE package."""

from typing import Any

import numpy as np
import pandas as pd
from fastkde import fastKDE
from scipy import interpolate

from empdens.base import AbstractDensity


def defaults():
    """Default parameters for the fastKDE."""
    return {}


class Interpolator:
    """Interpolates a grid of values in N-space to estimate values at arbitrary points."""

    def __init__(
        self,
        grid: np.ndarray,
        axes: list[np.ndarray],
        params: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the interpolator with a grid of values and the axes of the grid.

        Args:
            grid: This is an n-dimensional function surface.
            axes: Position labels for each dimension of `grid`.
            params: Additional named arguments to be passed to `scipy.interpolate.griddata` to control how the
                interpolation is done.
        """
        self.params = {}
        if params is not None:
            assert isinstance(params, dict)
            self.params = params
        self.values = grid.ravel()
        idx = np.meshgrid(*[np.arange(len(ax)) for ax in axes])
        indices = [ix.ravel() for ix in idx]
        self.positions = np.vstack([a[i] for a, i in zip(axes, indices)]).T

    def interpolate(self, x: np.ndarray) -> np.ndarray:
        """Interpolate the values at arbitrary points in N-space.

        Args:
            x: Points in the original N-space at which to estimate values.

        Returns:
            Interpolated values at the positions `x` based on the `self.values` observed over the `self.grid`.
        """
        interp = interpolate.griddata(
            points=self.positions,
            values=self.values,
            xi=x,
            **self.params,
        )
        if len(interp.shape) > 1:
            assert interp.shape[1] == 1, "Interpolation should return a single column"
            interp = interp[:, 0]
        # # The interpolator is evidently incapable of extrapolation; we get null values for points outside the grid.
        # exmax = (x > self.positions.max(axis=0)).max(axis=1)
        # exmin = (x < self.positions.min(axis=0)).max(axis=1)
        # assert not (set(np.where(np.isnan(interp))[0]) - set(np.where(exmax | exmin)[0]))
        # So let's fill null values with half the minimum value
        interp[np.isnan(interp)] = np.nanmin(interp) / 2
        return interp


class FastKDE(AbstractDensity):
    """Wrapper for fastKDE package."""

    def __init__(self, params: dict[str, Any] | None = None) -> None:
        """Initialize the fastKDE model.

        Args:
            params: Additional named arguments to be passed to `fastkde.fastKDE
        """
        super().__init__()
        self.params = defaults()
        if params is not None:
            self.params.update(params)

    def train(self, df: pd.DataFrame) -> None:
        """Train the density estimator on the given data.

        Currently throws a warning:
            https://bitbucket.org/lbl-cascade/fastkde/issues/5/using-a-non-tuple-sequence-for

        Args:
            df: Numeric features.
        """
        assert isinstance(df, pd.DataFrame), "FastKDE requires a pandas DataFrame"  # noqa: F821
        fkde = fastKDE.pdf(*[df[col].to_numpy() for col in df], **self.params)
        grid = fkde.data
        axes = [getattr(fkde, dim_name).to_numpy() for dim_name in reversed(fkde.dims)]
        self.interpolator = Interpolator(grid, axes)

    def density(self, X: pd.DataFrame) -> np.ndarray:
        """Estimate the density at the given points."""
        return self.interpolator.interpolate(X.to_numpy())



"""LightGBM classifier."""

import copy
import os
import pickle
import tempfile
from time import time
from typing import Any

import lightgbm as lgb
import numpy as np
import pandas as pd
from psutil import cpu_count

from empdens.classifiers.base import AbstractLearner


class Lgbm(AbstractLearner):
    """LightGBM classifier."""

    def __init__(
        self,
        params: dict[str, Any] | None = None,
        verbose: bool = False,
    ) -> None:
        """Initializes the LightGBM model."""
        super().__init__(params, verbose)
        self.nround = self.params.pop("num_boost_round")

    def default_params(self):
        """Returns the default parameters for the LightGBM model.

        Returns:
            dict: A dictionary containing the default parameters.
        """
        return {
            "task": "train",
            "boosting_type": "gbdt",
            "objective": "xentropy",
            "learning_rate": 0.07,
            "num_leaves": 20,
            "min_data_in_leaf": 20,
            "verbose": -1,
            "num_boost_round": 50,
            "num_threads": cpu_count(logical=False),
        }

    def _parse_categoricals(self):
        """Parses the categorical features for the LightGBM model."""
        if self.categorical_features is None:
            self.categoricals = "auto"
        else:
            self.categoricals = self.categorical_features
            assert all([c in self.features for c in self.categoricals])

    def as_lgb_data(self, data):
        """Converts the input data to a LightGBM Dataset.

        Args:
            data (empdens.data.Data): The input data.

        Returns:
            lgb.Dataset: The LightGBM Dataset.
        """
        self.features = data.X.columns.tolist()
        self.categorical_features = data.X.select_dtypes(include=["category"]).columns.tolist()
        self._parse_categoricals()
        return lgb.Dataset(
            data.X,
            data.y,
            feature_name=self.features,
            categorical_feature=self.categoricals,  # pyright: ignore[reportArgumentType]
        )

    def train(self, data):
        """Trains the LightGBM model.

        Args:
            data (empdens.data.Data): The input data.
        """
        t0 = time()
        ld = self.as_lgb_data(data)
        self.bst = lgb.train(
            params=copy.deepcopy(self.params),
            train_set=ld,
            num_boost_round=self.nround,
        )
        tdiff = str(round(time() - t0))
        self.vp("LightGBM training took " + tdiff + " seconds")

    def predict(self, X) -> np.ndarray:
        """Predicts the target values using the trained LightGBM model.

        Args:
            X (pd.DataFrame): The input features.

        Returns:
            np.ndarray: The predicted target values.
        """
        preds = self.bst.predict(X)
        assert isinstance(preds, np.ndarray), "LightGBM predictions must be a numpy array"
        return preds

    def freeze(self):
        """Serializes the trained LightGBM model to a binary attribute."""
        assert self.bst is not None
        _, filename = tempfile.mkstemp()
        self.bst.save_model(filename)
        with open(filename, "rb") as file:
            self.bst_binary = file.read()
        os.remove(filename)

    def thaw(self):
        """Deserializes the LightGBM model from the binary attribute."""
        assert hasattr(self, "bst_binary")
        assert self.bst_binary is not None
        self.bst = pickle.loads(self.bst_binary)

    def importance(self):
        """Returns the feature importance of the trained LightGBM model.

        Returns:
            pd.DataFrame: A DataFrame containing the feature importance.
        """
        return (
            pd.DataFrame({"feature": self.features, "gain": self.bst.feature_importance(importance_type="gain")})
            .sort_values("gain", ascending=False)
            .reset_index(drop=True)
        )



"""Wrapper for sklearn's KernelDensity class."""

from typing import Any

import numpy as np
import pandas as pd
from sklearn.neighbors import KernelDensity

from empdens.base import AbstractDensity


def defaults():
    """Default parameters for the KDE."""
    return {}


class SklearnKDE(AbstractDensity):
    """Wrapper for sklearn's KernelDensity class."""

    def __init__(self, params: dict[str, Any] | None = None) -> None:
        """Initialize the KDE model.

        Args:
            params: Additional named arguments to be passed to `sklearn.neighbors.KernelDensity`.
        """
        super().__init__()
        self.params = defaults()
        if params is not None:
            self.params.update(params)

    def train(self, df: pd.DataFrame) -> None:
        """Train the KDE model.

        Args:
            df: Data to train the model on.
        """
        self.kde = KernelDensity(**self.params)
        _ = self.kde.fit(df.to_numpy())

    def density(self, X: pd.DataFrame) -> np.ndarray:
        """Estimate the density of the data at the given points."""
        log_dens = self.kde.score_samples(X.to_numpy())
        return np.exp(log_dens)



"""The AbstractDensity class, which is the base class."""

from abc import ABC, abstractmethod

import pandas as pd


class AbstractDensity(ABC):
    """A class of method associated with a density.

    'density' is the only mandatory method, but this includes placeholders for several closely-related methods
    that often would expected in uses cases involving densities.

    Conceptually 'pdf' and 'density' are the same thing, but they differ
    in usage: pdf accepts a single 1-d vector representing a single point,
    while 'density' accepts a pandas DataFrame where each row is a point
    """

    def __init__(self, verbose: bool = False):
        """Initialize the AbstractDensity class."""
        self.verbose = verbose

    def train(self, df: pd.DataFrame) -> None:
        """A method for defining or updating the self.density function base on data."""
        raise Exception("Not yet implemented")

    @abstractmethod
    def density(self, X):
        """Return the density for each row of the pandas DataFrame X as a numpy array."""
        raise Exception("Not yet implemented")

    def predict(self, *args, **kwargs):
        """Prediction alias for density."""
        return self.density(*args, **kwargs)

    def rvs(self, n: int) -> pd.DataFrame:
        """Returns n samples from the space."""
        raise Exception("Not yet implemented")

    def pdf(self, X):
        """Evaluate the density function at a single point."""
        raise Exception("Single-point evaluation not yet supported, but see self.density")

    def vp(self, string):
        """A print function that respects self.verbose."""
        if self.verbose:
            print(string)

    def _validate_data(self, data):
        try:
            assert isinstance(data, pd.DataFrame)
        except Exception:
            raise Exception("the data needs to be a pandas.DataFrame")
        try:
            assert isinstance(data.columns[0], str)
        except Exception:
            raise Exception("the data column names need to be strings, not " + str(type(data.columns[0])))



"""Piecewise uniform density estimator."""

import numpy as np
import pandas as pd
import shmistogram as shmist
from scipy import stats
from shmistogram.binners.bayesblocks import BayesianBlocks
from shmistogram.tabulation import SeriesTable

from empdens.base import AbstractDensity
from empdens.models.multinomial import Multinomial


class PiecewiseUniform(AbstractDensity):
    """Adaptive-width histogram density estimator.

    Uses a shmistogram (https://github.com/zkurtz/shmistogram) to separates data
    into
    - 'loner' points (or modes), represented with a multinomial distribution as point masses
    - 'crowd' points, approximated by a standard piecewise uniform distribution
    """

    def __init__(
        self,
        alpha: float | None = None,
        loner_min_count: int = 20,
        binner=None,
        verbose: int = 0,
    ) -> None:
        """Initialize the piecewise uniform density estimator.

        Args:
            alpha: the proportion of loner points
            loner_min_count: minimum number of loner points
            binner: a binner object from shmistogram
            verbose: level of verbosity
        """
        super().__init__()
        self.alpha = alpha
        self.loner_min_count = loner_min_count
        if binner is None:
            self.binner = BayesianBlocks({"sample_size": 10000})
        self.verbose = verbose

    def _uniform(self, bin):
        return stats.uniform(bin.lb, bin.ub - bin.lb)

    def _train_loners(self, loners: SeriesTable) -> None:
        if loners.n == 0:
            self.multinomial = None
        else:
            m = Multinomial()
            m.train(loners.df)
            self.multinomial_df = self.loner_crowd_shares[0] * m.df[["density"]]
            self.multinomial = m

    def _set_null_crowd(self):
        self.crowd_uniforms = []
        self.crowd_lookup = pd.DataFrame({"xval": [], "density": []})

    def _train_crowd(self, crowd_bins):
        self.crowd_bins = crowd_bins
        # A multinomial distribution determines which bin to draw from
        if crowd_bins is None:
            self._set_null_crowd()
        elif crowd_bins.shape[0] == 0:
            self._set_null_crowd()
        else:
            self.crowd_multinom = Multinomial()
            series = self.crowd_bins.rename(columns={"freq": self.name})[self.name]
            self.crowd_multinom.train(df=series.to_frame())
            self.crowd_uniforms = [self._uniform(row) for _, row in self.crowd_bins.iterrows()]
            # A density lookup for each member of the crowd (assuming asof backward merge)
            crowd_share = self.loner_crowd_shares[1]
            lookup = self.crowd_bins[["lb"]].rename(columns={"lb": "xval"})
            lookup["density"] = crowd_share * self.crowd_bins.rate / self.crowd_bins.freq.sum()
            self.crowd_lookup = pd.concat(
                [lookup, pd.DataFrame({"xval": self.crowd_bins[["ub"]].max().values, "density": [np.nan]})], sort=True
            )

    def train(self, df: pd.DataFrame) -> None:
        """Train a model.

        Args:
            df: A single-column data frame.
        """
        assert df.shape[1] == 1, "Only one-dimensional data is supported"
        self.name = df.columns[0]
        series = df[self.name]
        shm = shmist.Shmistogram(series, binner=self.binner)
        self.loner_crowd_shares = shm.loner_crowd_shares
        # Loners
        self._train_loners(shm.loners)
        # Crowd
        self._train_crowd(shm.bins)
        # Define density for out-of-sample obs as half the min observed density:
        mmin = np.inf
        if self.multinomial is not None:
            mmin = self.multinomial_df.density.min()
        self.oos_density = min(mmin, self.crowd_lookup.density.min(), 1 / shm.n_obs) / 2

    def density(self, X: pd.DataFrame) -> np.ndarray:
        """Compute the density function on each row of X."""
        if X.shape[1] > 1:
            raise ValueError("Only one-dimensional data is supported")
        x_series = X.iloc[:, 0]
        # Identify the unique values for which densities are needed
        ref = pd.DataFrame({"xval": x_series.unique()})
        if self.multinomial is not None:
            # Look up each loner in the multinomial levels; those with no match will be
            #   treated as members of the crowd
            ref = ref.merge(self.multinomial_df, left_on="xval", right_index=True, how="left")
            is_crowd = np.isnan(ref.density)
            ref_loners = ref[~is_crowd].copy()
            ref_crowd = ref[is_crowd].drop("density", axis=1)
        else:
            ref_loners = pd.DataFrame(columns=pd.Index(["xval"]))
            ref_crowd = ref
        ref_crowd = ref_crowd.sort_values(by="xval")  # pyright: ignore
        ref_crowd = ref_crowd.reset_index(drop=True)
        if self.crowd_lookup.xval.dtype == "float64":
            ref_crowd.xval = ref_crowd.xval.astype("float64")
        ref_crowd_roll = pd.merge_asof(ref_crowd, self.crowd_lookup, on="xval")
        assert isinstance(ref_crowd_roll, pd.DataFrame), "ref_crowd_roll is None"
        assert isinstance(ref_loners, pd.DataFrame), "ref_loners is None"
        frames = [ref_loners, ref_crowd_roll]
        nonempty_frames = [frame for frame in frames if not frame.empty]
        final_ref = pd.concat(nonempty_frames, axis=0)
        final_ref["density"] = final_ref.density.fillna(self.oos_density)
        xdf = pd.DataFrame({"xval": x_series.to_numpy(), "order": range(len(X))})
        result = final_ref.merge(xdf, right_on="xval", left_on="xval", how="right").sort_values("order")
        assert result.shape[0] == len(X), "Mismatch in length"
        return result.density.to_numpy()

    def rvs(self, n: int) -> pd.DataFrame:
        """Simulate n draws."""
        # Flip coins to determine number of loners versus crowd
        n_loners = stats.binom.rvs(n=n, p=self.loner_crowd_shares[0], size=1)
        n_crowd = n - n_loners
        # Sample the loners
        if n_loners > 0:
            assert self.multinomial is not None, "Multinomial not defined"
            loners = self.multinomial.rvs(n_loners)[self.name].to_numpy()
        else:
            loners = np.array([])
        # Sample the crowd
        if n_crowd > 0:
            assert self.crowd_multinom is not None, "Crowd multinomial not defined"
            values_df = self.crowd_multinom.rvs(n_crowd)
            assert values_df.columns.to_list() == [self.name]
            bins = values_df[self.name].value_counts()
            crowd_list = [self.crowd_uniforms[k].rvs(bins.iloc[k]) for k in range(len(bins))]
            crowd = np.array([x for y in crowd_list for x in y])
        else:
            crowd = np.array([])
        # Shuffle
        data = np.concatenate((loners, crowd))
        np.random.shuffle(data)
        return pd.Series(data, name=self.name).to_frame()



"""Wrapper for sklearn's IsolationForest."""

from typing import Any

import numpy as np
import pandas as pd
from sklearn.ensemble import IsolationForest

from empdens.base import AbstractDensity


def defaults():
    """Default parameters for the IsolationForest."""
    return {"contamination": "auto"}


class SklearnIsolationForest(AbstractDensity):
    """Wrapper for sklearn's IsolationForest."""

    def __init__(self, params: dict[str, Any] | None = None) -> None:
        """Initialize the IsolationForest model.

        Args:
            params: Additional named arguments to be passed to `sklearn.ensemble.IsolationForest`.
        """
        super().__init__()
        self.params = defaults()
        if params is not None:
            self.params.update(params)

    def train(self, df: pd.DataFrame) -> None:
        """Train the model.

        Args:
            df: Data to train the model on.
        """
        self.forest = IsolationForest(**self.params)  # pyright: ignore
        _ = self.forest.fit(df.to_numpy())

    def density(self, X: pd.DataFrame) -> np.ndarray:
        """Estimate the density of the data at the given points.

        Args:
            X: Data to estimate the density over.
        """
        # TODO: explain how this is kinda sorta getting a density
        dens = 1 + self.forest.score_samples(X.to_numpy())
        return dens



"""Base class for all classifiers in the empdens package."""

import copy
from abc import ABC, abstractmethod
from typing import Any

import numpy as np


class AbstractLearner(ABC):
    """Base class for all classifiers in the empdens package."""

    def __init__(self, params: dict[str, Any] | None = None, verbose=False):
        """Initialize the base class for all classifiers in the empdens package."""
        self.params = self.default_params()
        if params is not None:
            self.params.update(copy.deepcopy(params))
        self.verbose = verbose

    @abstractmethod
    def default_params(self) -> dict[str, Any]:
        """Return default parameters for the classifier."""
        pass

    @abstractmethod
    def train(self, data):
        """Train the model."""
        pass

    @abstractmethod
    def predict(self, X) -> np.ndarray:
        """Predict on new data."""
        pass

    def vp(self, string):
        """Prints a string if verbose is True."""
        if self.verbose:
            print(string)



"""Classifier-adjusted density estimation."""

import numpy as np
import pandas as pd
from pandahandler import Schema, categorize_non_numerics
from sklearn import metrics

from empdens import models
from empdens.base import AbstractDensity
from empdens.classifiers.base import AbstractLearner
from empdens.classifiers.lightgbm import Lgbm
from empdens.data import CadeData


def auc(df):
    """Compute the area under the ROC curve."""
    fpr, tpr, _ = metrics.roc_curve(df.truth.values, df.pred.values, pos_label=1)
    return metrics.auc(fpr, tpr)


class Cade(AbstractDensity):
    """Classifier-adjusted density estimation.

    Based on https://pdfs.semanticscholar.org/e4e6/033069a8569ba16f64da3061538bcb90bec6.pdf
    """

    # A soft target for the number of instances to simulate when `sim_size` is "auto"
    simulation_size_attractor = 10000

    def __init__(
        self,
        initial_density: AbstractDensity | None = None,
        classifier: AbstractLearner | None = None,
        sim_size: str | int = "auto",
        verbose: bool = False,
    ):
        """Initialize the classifier-adjusted density estimation model.

        Args:
            initial_density: A naive density model to use as the basis for CADE estimation. CADE uses this as a
                reference such that the final estimated density is a product of the initial density and the classifier
                adjustment. If None, defaults to a JointDensity model.
            classifier: A classifier to use for the adjustment. If None, defaults to a LightGBM classifier.
            sim_size: The number of synthetic samples to simulate. If "auto", the simulation size is set as the
                geometric mean between the data size and `simulation_size_attractor`. If a positive number less than
                100, the simulation size is `round(sim_size)*df.shape[0]`. If `sim_size` is greater than or equal to
                100, the simulation size is `round(sim_size)`.
            verbose: Whether to print diagnostic information.
        """
        super().__init__()
        self.initial_density = initial_density or models.JointDensity(verbose=verbose)
        self.classifier = classifier or Lgbm()
        self.sim_size = sim_size
        self.verbose = verbose

    def compute_simulation_size(self, df):
        """Determine the number of synthetic data samples to simulate.

        If self.sim_size is 'auto', sets the simulation size as the geometric mean
        between the data size and self.simulation_size_attractor

        If self.sim_size is a positive number less than 100, simulation size is
        round(self.sim_size)*df.shape[0]

        Finally, if self.sim_size >= 100, simulation size is round(self.sim_size)
        """
        n_real = df.shape[0]
        if isinstance(self.sim_size, str):
            assert self.sim_size == "auto"
            sim_n = np.sqrt(n_real * self.simulation_size_attractor)
        elif self.sim_size < 100:
            assert self.sim_size > 0
            sim_n = round(self.sim_size * n_real)
            if sim_n < 10:
                raise Exception("Simulation size is very small. Consider using a larger value of sim_size")
        else:
            sim_n = round(self.sim_size)
        self.sim_rate = sim_n / df.shape[0]
        return int(sim_n)

    def _diagnostics(self, x, truth):
        val_df = pd.DataFrame({"pred": self.classifier.predict(x), "truth": truth})
        self.diagnostics = {
            "val_df": val_df,
            "auc": auc(val_df),
        }

    def train(self, df: pd.DataFrame, diagnostics: bool = False):
        """Model the density of the data.

        :param df: (pandas DataFrame)
        """
        df = categorize_non_numerics(df)
        self.schema = Schema.from_df(df)
        self.vp(f"Training a generative density model on {len(df)} samples")
        self.initial_density.train(df)
        sim_n = self.compute_simulation_size(df)
        self.vp(f"Simulating {sim_n} fake samples from the model and join it with the real data")
        sim_df = self.initial_density.rvs(sim_n)
        sim_df = self.schema(sim_df)
        xdf = pd.concat([df, sim_df], axis=0)
        assert isinstance(xdf, pd.DataFrame), "CadeData.X requires a pandas DataFrame"
        partially_synthetic_data = CadeData(
            X=xdf,
            y=np.concatenate([np.ones(df.shape[0]), np.zeros(sim_n)]),
        )
        self.vp("Train the classifier to distinguish real from fake")
        self.classifier.train(partially_synthetic_data)
        if diagnostics:
            self._diagnostics(partially_synthetic_data.X, partially_synthetic_data.y)
        if self.verbose:
            if not hasattr(self, "diagnostics"):
                self._diagnostics(partially_synthetic_data.X, partially_synthetic_data.y)
            AUROC = str(round(self.diagnostics["auc"], 3))
            print("In-sample, the classifier had AUROC = " + AUROC)

    def density(self, X):
        """Predict the density at new points.

        Apply equation 2.1 in https://pdfs.semanticscholar.org/e4e6/033069a8569ba16f64da3061538bcb90bec6.pdf

        Args:
            X: Data frame matching the schema of the training data.
        """
        X = self.schema(X)
        # Initial density estimate
        synthetic_dens = self.initial_density.density(X)
        # Classifier adjustment factor
        p_real = self.classifier.predict(X)
        odds_real = p_real / (1 - p_real)
        classifier_adjustment = self.sim_rate * odds_real
        return synthetic_dens * classifier_adjustment



"""Model a single categorical feature."""

from typing import Any

import numpy as np
import pandas as pd
from shmistogram.tabulation import SeriesTable, tabulate

from empdens.base import AbstractDensity


class Multinomial(AbstractDensity):
    """Model a single categorical feature."""

    def _density(self):
        reg_counts = self.df.n_obs.to_numpy() + 1
        self.df["density"] = reg_counts / reg_counts.sum()
        # Assign a tiny prob for never-before observed values of this multinomial:
        self.out_of_sample_dens = 1 / (2 * self.df.n_obs.sum())

    def _train_empirically(self, series: SeriesTable | pd.Series):
        if isinstance(series, SeriesTable):
            st = series
            assert st.df.shape[0] > 0
            self.name = st.name
        else:
            assert isinstance(series, pd.Series), f"Input must be a pandas Series, got {type(series)}"
            self.name = series.index.name if series.name == "n_obs" else series.name
            assert len(series) > 0, "Empty series"
            st = tabulate(series)
        self.df = st.df

    def _train_by_accepting_params(self, counts, values=None):
        self.df = pd.DataFrame({"n_obs": counts})
        if values is not None:
            self.df.index = values

    def train(self, df: pd.DataFrame, counts=None, values=None):
        """Specify at least series or counts but not both.

        :param series: (pandas.Series or SeriesTable of integers
        :param counts: numpy 1-d array of counts corresponding to
        each entry of `values`
        :param values: numpy 1-d array; integers representing each multinomial outcome;
        ignored if `counts` is None.
        """
        if df.shape[1] > 1:
            raise Exception("Only one-dimensional data is supported")
        series = df[df.columns[0]]
        # if series is not None:
        assert isinstance(series, pd.Series), "Input must be a pandas Series"
        self._train_empirically(series)
        # else:
        #     assert counts is not None
        #     assert counts.sum() > 0
        #     self._train_by_accepting_params(counts, values=values)
        self._density()

    def point_density(self, item: Any) -> np.ndarray:
        """Compute the density for an individual value. TODO: should not be for individual value."""
        raise NotImplementedError

    def density(self, X: pd.DataFrame) -> np.ndarray:
        """Fast density computation for a list of values.

        Args:
            X: A data frame with one column of values.
        """
        if X.shape[1] > 1:
            raise Exception("Only one-dimensional data is supported")
        series = X[X.columns[0]]
        assert isinstance(series, pd.Series)
        df = (
            pd.DataFrame({"levels": series, "idx": range(len(series))})
            .merge(
                pd.DataFrame({"levels": self.df.index.values, "density": self.df.density.values}),
                on="levels",
                how="left",
            )
            .sort_values("idx")
        )
        return df.density.fillna(self.out_of_sample_dens).to_numpy()

    def rvs(self, n: int = 1) -> pd.DataFrame:
        """Randomly sample from the multinomial distribution."""
        values = np.random.choice(a=self.df.index.to_numpy(), size=n, p=self.df.density.to_numpy(), replace=True)
        return pd.Series(values, name=self.name).to_frame()



"""Decision tree density estimation."""

import numpy as np

from empdens.base import AbstractDensity

# TODO: generalize to handle categoricals


def log_node_volume(node) -> float:
    """Compute the log-volume of a node in terms of its bounds."""
    diffs = node.bounds[1] - node.bounds[0]
    ldiffs = np.log(diffs)
    return np.sum(ldiffs)


def node_volume(node):
    """Compute the volume of a node in terms of its bounds."""
    lv = log_node_volume(node)
    return np.exp(lv)


def split_bounds(bounds, feature_idx, value) -> dict[str, np.ndarray]:
    """Split the bounds of a node along a feature.

    Args:
        bounds: Two rows, with upper and lower bounds for each feature.
        feature_idx: Index of feature to split on.
        value: Value to split on.
    """
    bounds_left = np.copy(bounds)
    bounds_right = np.copy(bounds)
    bounds_left[1, feature_idx] = value
    bounds_right[0, feature_idx] = value
    return {"left": bounds_left, "right": bounds_right}


class Node:
    """Store the attributes of a node.

    - bounds
    - if a leaf, the list of indices of training data contained
    - if not a leaf, the split that defines its children.
    """

    def __init__(self, bounds, members) -> None:
        """Initialize the node.

        :param bounds: (np.array) two rows, with upper and lower bounds for each feature
        :param members: (numpy integer array) indices of training data falling in this node
        """
        self.bounds = bounds
        self.members = members
        # until we decide to split it, this is simply a leaf
        self.leaf = True
        self.left = None
        self.right = None

    def split(
        self,
        feature_idx: int,
        value: float,
        left_members: np.ndarray,
        right_members: np.ndarray,
    ) -> None:
        """Split the node into two children.

        Args:
            feature_idx: (int) index of the feature to split on
            value: (float) value to split on
            left_members: (numpy integer array) indices of training data falling in the left child
            right_members: (numpy integer array) indices of training data falling in the right child
        """
        child_bounds = split_bounds(self.bounds, feature_idx, value)
        self.left = Node(bounds=child_bounds["left"], members=left_members)
        self.right = Node(bounds=child_bounds["right"], members=right_members)
        self.members = None
        self.leaf = False


class Tree:
    """A decision tree for density estimation."""

    def __init__(self, df):
        """Initialize the tree."""
        self.n_features = df.shape[1]
        self.n_points = df.shape[0]
        self._plant_me(df)
        while True:
            self._search_split(df)

    def _plant_me(self, df):
        bounds = np.array([df.min(axis=0), df.max(axis=0)])
        self.idx = list(range(self.n_points))
        self.root = Node(bounds, members=list(range(self.n_points)))
        self.leaves = [self.root]

    def _search_split(self, df):
        raise NotImplementedError("Not clear what the plan was for this.")


class TreeDensity(AbstractDensity):
    """A decision-tree density."""

    def __init__(self, verbose=False):
        """Initialize the decision tree density model."""
        super().__init__(verbose=verbose)

    def train(self, df):
        """Model the density of the data.

        :param df: (pandas DataFrame)
        """
        self._validate_data(df)
        self.vp("Training a decision tree density model on " + str(df.shape[0]) + " samples")
        self.tree = Tree(df)

    def density(self, X):
        """Predict the density at new points.

        Apply equation 2.1 in https://pdfs.semanticscholar.org/e4e6/033069a8569ba16f64da3061538bcb90bec6.pdf

        :param X: (pd.DataFrame or numpy array) Must match the exact column order of the `df`
            argument that was passed to self.train
        """
        self._validate_data(X)
        raise NotImplementedError("Not clear what the plan was for this.")



"""Model a joint density of multiple features."""

from typing import Any

import numpy as np
import pandas as pd

from empdens.base import AbstractDensity
from empdens.models.multinomial import Multinomial
from empdens.models.piecewise_uniform import PiecewiseUniform


class JointDensity(AbstractDensity):
    """Model a joint density of multiple features."""

    def __init__(self, numeric_params: dict[str, Any] | None = None, verbose: bool = False) -> None:
        """Initialize the joint density model."""
        super().__init__(verbose=verbose)
        self.multinom_class = Multinomial
        self.numeric_class = PiecewiseUniform
        self.numeric_params = numeric_params

    def _fit_categorical(self, series: pd.Series) -> Multinomial:
        model = self.multinom_class()
        model.train(series.to_frame())
        return model

    def _fit_continuous(self, values: pd.Series) -> PiecewiseUniform:
        params = {}
        if self.numeric_params is not None:
            params = self.numeric_params
        model = self.numeric_class(verbose=self.verbose - 1, **params)
        model.train(values.to_frame())
        return model

    def _fit_univarite(self, series: pd.Series) -> Multinomial | PiecewiseUniform:
        msg = "Fitting univariate density on " + str(series.name) + " as "
        if isinstance(series.dtype, pd.CategoricalDtype):
            self.vp(msg + "categorical")
            return self._fit_categorical(series)
        else:
            self.vp(msg + "continuous")
            return self._fit_continuous(series)

    def train(self, df: pd.DataFrame) -> None:
        """Train the joint density model on a DataFrame."""
        self.columns = df.columns
        self.univariates = {v: self._fit_univarite(df[v]) for v in self.columns}  # pyright: ignore

    def density(self, X: pd.DataFrame, log: bool = False):
        """Compute the density for an individual value."""
        assert isinstance(X, pd.DataFrame)
        assert all(X.columns == self.columns)
        df_log_univariate = pd.DataFrame(
            {v: np.log(self.univariates[v].density(X[[v]])) for v in self.columns},  # pyright: ignore
        )
        log_dens = df_log_univariate.sum(axis=1).to_numpy()
        if log:
            return log_dens
        return np.exp(log_dens)

    def rvs(self, n):
        """Generate n samples from the fitted distribution."""
        if not hasattr(self, "univariates"):
            raise Exception("Call `train` before you call `rvs`")
        samples = [self.univariates[col].rvs(n) for col in self.columns]
        return pd.concat(samples, axis=1)



"""Bivariate data simulators."""

import numpy as np
import pandas as pd
from scipy import stats

from empdens.base import AbstractDensity


class Zena(AbstractDensity):
    """A simple bivariate data simulator.

    This use two independent distributions, a Gaussian and a triangular distribution.
    """

    def __init__(self):
        """Initialize the bivariate data simulator."""
        super().__init__()
        self.gauss = stats.truncnorm(-2, 4)
        self.triang = stats.triang(0, 0, 3)

    def rvs(self, n: int) -> pd.DataFrame:
        """Simulate draws from the joint distribution."""
        return pd.DataFrame({"gaussian": self.gauss.rvs(size=n), "triangular": self.triang.rvs(size=n)})

    def density(self, X: pd.DataFrame) -> np.ndarray:
        """Calculate the density at the given points."""
        points = X.to_numpy()
        gauss_pdf = self.gauss.pdf  # pyright: ignore
        triang_pdf = self.triang.pdf  # pyright: ignore
        return np.array([gauss_pdf(p[0]) * triang_pdf(p[1]) for p in points])



"""Simulators for univariate densities."""

import numpy as np
import pandas as pd
from scipy import stats

from empdens.base import AbstractDensity


class Multinomial:
    """A multinomial random variable."""

    def __init__(self, probs: np.ndarray) -> None:
        """Define a multinomial random variable object.

        Args:
            probs: The probability of each class, with classes indexed as 0 to len(probs)-1
        """
        assert isinstance(probs, np.ndarray)
        self.idx = list(range(len(probs)))
        self.probs = probs / probs.sum()

    def rvs(self, n: int) -> np.ndarray:
        """Simulate n draws from the multinomial distribution."""
        return np.random.choice(a=self.idx, size=n, p=self.probs, replace=True)

    def density(self, X: pd.DataFrame) -> np.ndarray:
        """Calculate the density at the given points."""
        assert X.shape[1] == 1
        points = X.to_numpy()[:, 0]
        return np.array([self.probs[k] for k in points])


class BartSimpson(AbstractDensity):
    """The "claw".

    As in https://projecteuclid.org/download/pdf_1/euclid.aos/1176348653;
    renamed as in http://www.stat.cmu.edu/~larry/=sml/densityestimation.pdf.
    """

    def __init__(self) -> None:
        """Initialize the Bart Simpson density."""
        super().__init__()
        # Guassians to mix over
        self.gaussians = [
            stats.norm(),
            stats.norm(loc=-1, scale=0.1),
            stats.norm(loc=-0.5, scale=0.1),
            stats.norm(loc=0, scale=0.1),
            stats.norm(loc=0.5, scale=0.1),
            stats.norm(loc=1, scale=0.1),
        ]
        # Mixing weights
        self.multinomial = Multinomial(probs=np.array([0.5] + [0.1] * 5))

    def rvs(self, n: int) -> pd.DataFrame:
        """Simulate n draws."""
        idxs = self.multinomial.rvs(n)
        values, counts = np.unique(idxs, return_counts=True)
        samples = [self.gaussians[values[k]].rvs(counts[k]) for k in range(len(values))]
        samples = [v for sublist in samples for v in sublist]
        np.random.shuffle(samples)
        return pd.DataFrame({"bart_simpson": samples})

    def density(self, X: pd.DataFrame) -> np.ndarray:
        """Calculate the density at the given points."""
        assert X.shape[1] == 1
        points = X.to_numpy()[:, 0]
        pdf_funcs = [gauss.pdf for gauss in self.gaussians]  # pyright: ignore
        return np.column_stack([func(points) for func in pdf_funcs]) @ self.multinomial.probs

    def plot(self, n: int = 200, xlims: list[float | int] = [-2, 2]) -> None:
        """Plot the density."""
        dfg = pd.DataFrame({"x": np.linspace(xlims[0], xlims[1], n)})
        dfg["generative density"] = self.density(dfg)
        ax = dfg.plot(x="x", y="generative density")
        ax.get_legend().remove()



"""Data structures."""

from importlib import resources
from pathlib import Path

import numpy as np
import pandas as pd


class CadeData(object):
    """A standardized data format for empdens.cade.Cade."""

    def __init__(self, X: pd.DataFrame, y: pd.Series | np.ndarray) -> None:
        """Initialize the CadeData object.

        Args:
            X: The feature matrix.
            y: The target vector.
        """
        assert len(y.shape) == 1
        assert len(y) == X.shape[0]
        self.X = X
        self.y = y


def load_Japanese_vowels_data() -> pd.DataFrame:
    """Data downloaded from http://odds.cs.stonybrook.edu/japanese-vowels-data/."""
    path = resources.files("empdens.resources.data").joinpath("japanese_vowels.csv")
    assert isinstance(path, Path)
    return pd.read_csv(path)


def load_SHAP_census_data():
    """Loads the 'adults' dataset available in SHAP.

    This borrows a few SHAP file parsing code snippets, https://github.com/slundberg/shap/blob/master/shap/datasets.py.
    """
    try:
        pass
    except Exception:
        raise Exception("Do `pip install shap` and try again")
    dtypes = [
        ("Age", "float32"),
        ("Workclass", "category"),
        ("fnlwgt", "float32"),
        ("Education", "category"),
        ("Education-Num", "float32"),
        ("Marital Status", "category"),
        ("Occupation", "category"),
        ("Relationship", "category"),
        ("Race", "category"),
        ("Sex", "category"),
        ("Capital Gain", "float32"),
        ("Capital Loss", "float32"),
        ("Hours per week", "float32"),
        ("Country", "category"),
        ("Target", "category"),
    ]
    df = pd.read_csv(
        "https://github.com/slundberg/shap/raw/master/data/adult.data",
        names=[d[0] for d in dtypes],
        na_values="?",
        dtype=dict(dtypes),
    )
    df = df[df.Country == " United-States"].copy()
    df.drop(["Country", "Education", "fnlwgt"], axis=1, inplace=True)
    df.rename({"Target": "Income"}, axis=1, inplace=True)
    return df



"""Evaluation of binary predictions."""

import numpy as np
from sklearn import metrics


class Binary(object):
    """A collection of metrics for the strength of association between two vectors in [0,1]."""

    def __init__(self, truth, pred):
        """Initialize the Binary object."""
        assert len(truth) == len(pred)
        assert isinstance(truth, np.ndarray)
        assert isinstance(pred, np.ndarray)
        self.truth = truth
        self.pred = pred
        self.metrics = {
            "auroc": self.AUROC
            # 'rank-order correlation': self.rank_order_correlation,
            # 'pearson correlation': self.pearson_correlation
        }

    def AUROC(self):
        """Area under the receiver-operator characteristic curve."""
        return metrics.roc_auc_score(y_true=self.truth, y_score=self.pred)



"""Evaluation of density estimators."""

import numpy as np
import pandas as pd


class Evaluation(object):
    """A collection of metrics of density estimation performance."""

    def __init__(self, estimators, truth=None):
        """Initialize the Evaluation object.

        Args:
            estimators: A dictionary of estimator predictions.
            truth: The true density values.
        """
        for est in estimators:
            assert isinstance(estimators[est], np.ndarray)
        self.estimators = estimators
        self.truth = truth
        if truth is not None:
            assert isinstance(truth, np.ndarray)
        self.metrics = {
            "mean_absolute_error": self.mean_absolute_error,
            "mean_squared_error": self.mean_squared_error,
            "rank-order correlation": self.rank_order_correlation,
            "pearson correlation": self.pearson_correlation,
            "mean density": self.mean_density,
        }

    def _assert_truth_available(self):
        """Raise an exception if truth is not available."""
        if self.truth is None:
            raise Exception("rank-order correlation can't be computed since you did not provide anything for `truth`")

    def rank_order_correlation(self, pred):
        """Spearman (i.e. rank-order) correlation of prediction against truth.

        Maximize me
        """
        self._assert_truth_available()
        s_pred = pd.Series(pred)
        t_pred = pd.Series(self.truth)
        return s_pred.corr(t_pred, method="spearman")

    def pearson_correlation(self, pred):
        """Spearman (i.e. rank-order) correlation of prediction against truth.

        Maximize me
        """
        self._assert_truth_available()
        s_pred = pd.Series(pred)
        t_pred = pd.Series(self.truth)
        return s_pred.corr(t_pred)

    def mean_absolute_error(self, pred):
        """Direct estimate of expectation of L1 loss."""
        self._assert_truth_available()
        return np.absolute(pred - self.truth).mean()

    def mean_squared_error(self, pred):
        """Direct estimate of expectation of L2 loss."""
        self._assert_truth_available()
        squared_errors = (pred - self.truth) ** 2
        return squared_errors.mean()

    def mean_density(self, pred):
        """The mean density; this is exp(-deviance).

        This is a sensible metric for some (but not all) density estimators
        """
        return np.mean(pred)

    def evaluate_estimator(self, name):
        """Evaluate the estimator with the given name."""
        est = self.estimators[name]
        return [fun(est) for name, fun in self.metrics.items()]

    def evaluate(self):
        """Evaluate all estimators."""
        estimators = self.estimators.keys()
        return pd.DataFrame({e: self.evaluate_estimator(e) for e in estimators}, index=pd.Index(self.metrics))
